from dataclasses import dataclass
from .calltype import CallType


@dataclass
class ServiceCallsDTO:
    on: CallType
    off: CallType
    pause: CallType | None = None
    resume: CallType | None = None
    update_current: CallType | None = None

    def __post_init__(self):
        if self.pause is None:
            self.pause = self.off
        if self.resume is None:
            self.resume = self.on



from dataclasses import dataclass, field
from datetime import datetime, date, timedelta


@dataclass
class DateTimeModel:
    _datetime: datetime = field(default_factory=datetime.now)
    _datetime_set: bool = False
    _date: date = field(default_factory=date.today)
    _date_set: bool = False
    _hour: int = 0
    _hour_set: bool = False
    _quarter: int = 0
    _quarter_set: bool = False

    def set_datetime(self, mock_dt: datetime):
        assert isinstance(mock_dt, datetime), "Datetime must be a datetime object"
        self._datetime = mock_dt.replace(minute=0, second=0, microsecond=0)
        self._datetime_set = True
        self._date = mock_dt.date()
        self._date_set = True
        self._hour = mock_dt.hour
        self._hour_set = True
        self._quarter = mock_dt.minute // 15
        self._quarter_set = True

    def set_date(self, mock_date: date):
        assert isinstance(mock_date, date), "Date must be a date object"
        self._date = mock_date
        self._date_set = True

    def set_hour(self, mock_hour: int):
        assert 0 <= mock_hour <= 23, "Hour must be between 0 and 23"
        self._hour = mock_hour
        self._hour_set = True

    def set_quarter(self, mock_quarter: int):
        assert 0 <= mock_quarter <= 3, "Quarter must be between 0 and 3"
        self._quarter = mock_quarter
        self._quarter_set = True

    def is_passed(self, datum, hour, quarter) -> bool:
        if datum == self.hdate_tomorrow:
            return False
        if self.hour > hour:
            return True
        elif self.hour == hour:
            if self.quarter > quarter:
                return True
        return False

    @property
    def dt(self) -> datetime:
        return (
            self._datetime
            if self._datetime_set
            else datetime.now().replace(minute=0, second=0, microsecond=0)
        )

    @property
    def hdate(self) -> date:
        return self._date if self._date_set else date.today()

    @property
    def hdate_tomorrow(self) -> date:
        return self.hdate + timedelta(days=1)

    @property
    def hour(self) -> int:
        return self._hour if self._hour_set else datetime.now().hour

    @property
    def quarter(self) -> int:
        return self._quarter if self._quarter_set else (datetime.now().minute // 15)



from dataclasses import dataclass, field
from datetime import date, datetime, time
from .hour_type import HourType
from .list_type import ListType
from .datetime_model import DateTimeModel


@dataclass
class HourPrice:
    dt: datetime = datetime.now()
    quarter: int = 0
    price: float = 0.0
    permittance: float = field(init=False)
    passed: bool = False
    hour_type: HourType = HourType.Regular
    list_type: ListType = ListType.Hourly

    @property
    def hour(self) -> int:
        return self.dt.hour

    def __post_init__(self):
        assert 0 <= self.quarter <= 3, "Quarter must be between 0 and 3"
        assert 0 <= self.dt.hour <= 23, "Hour must be between 0 and 23"
        self.permittance = 1.0 if self.hour_type == HourType.BelowMin else 0.0

    @staticmethod
    def set_hour_type(max_price, min_price, price):
        if price > max_price:
            return HourType.AboveMax
        elif price < min_price:
            return HourType.BelowMin
        return HourType.Regular

    def set_passed(self, dt: DateTimeModel):
        if dt.hdate > self.dt.date():
            self.passed = True
        elif dt.hdate == self.dt.date():
            if self.hour < dt.hour:
                self.passed = True
            elif all(
                [
                    self.hour == dt.hour,
                    self.quarter < dt.quarter,
                    self.list_type == ListType.Quarterly,
                ]
            ):
                self.passed = True
        else:
            self.passed = False



from __future__ import annotations
from typing import TYPE_CHECKING
from datetime import datetime, timedelta
from ..hoursselection_service_new.models.hour_price import HourPrice

if TYPE_CHECKING:
    from .hourselection_service import HourSelectionService
from typing import Tuple
from .models.max_min_model import MaxMinModel

MINIMUM_DIFFERENCE = 0.1


class MaxMinCharge:
    def __init__(self, service: HourSelectionService, min_price: float | None) -> None:
        self.model = MaxMinModel(min_price=min_price)  # type: ignore
        self.parent = service
        self.active: bool = False

    @property
    def average_price(self) -> float | None:
        if not self.active:
            return None
        return self.model.caluclate_average_price(
            self.model.input_hours, self.total_charge, self.parent.dtmodel.dt
        )

    @property
    def original_average_price(self) -> float | None:
        return self.model.caluclate_average_price(
            self.model.original_input_hours,
            self.original_total_charge,
            self.parent.dtmodel.dt,
        )

    @property
    def total_charge(self) -> float | None:
        if not self.active:
            return None
        return self.model.calculate_total_charge(
            self.model.input_hours, self.parent.dtmodel.dt
        )

    @property
    def original_total_charge(self) -> float:
        return self.model.calculate_total_charge(
            self.parent.future_hours, self.parent.dtmodel.dt
        )

    @property
    def non_hours(self) -> list:
        return [
            hp.dt
            for hp in self.model.input_hours
            if hp.permittance == 0 and hp.dt >= self.parent.dtmodel.dt
        ]

    @property
    def dynamic_caution_hours(self) -> dict:
        return {
            hp.dt: hp.permittance
            for hp in self.model.input_hours
            if 0 < hp.permittance < 1 and hp.dt >= self.parent.dtmodel.dt
        }

    async def async_allow_decrease(self, car_connected: bool | None = None) -> bool:
        if car_connected is not None:
            return all(
                [
                    # not car_connected,
                    len([v for v in self.model.input_hours if v.permittance > 0])
                    != 1,
                ]
            )
        return len([v for v in self.model.input_hours if v.permittance > 0]) != 1

    async def async_update(
        self,
        avg24,
        peak,
        max_desired: float,
        session_energy: float | None = None,
        car_connected: bool | None = None,
    ) -> None:
        # allow_decrease: bool = False
        if await self.async_allow_decrease(car_connected):
            # allow_decrease = True
            await self.async_setup(max_charge=peak)
        _session = session_energy or 0
        _desired = max_desired - _session
        _avg24 = round((avg24 / 1000), 1)
        self.model.expected_hourly_charge = peak - _avg24
        self.select_hours_for_charge(self.model.original_input_hours, _desired)

    def select_hours_for_charge(
        self, hours: list[HourPrice], desired_charge: float
    ) -> None:
        hours = [hour for hour in hours if hour.permittance != 0]
        total_charge = 0
        hours.sort(key=lambda x: x.price)
        for hour in hours:
            hour_charge = hour.permittance * self.model.expected_hourly_charge
            # print(
            #     f"test {hour.dt} with {hour.permittance}. Charge: {hour_charge}, total: {total_charge}"
            # )
            if total_charge + hour_charge <= desired_charge:
                total_charge += hour_charge
            elif total_charge + hour_charge > desired_charge:
                perm = max(min(desired_charge - total_charge, 1), 0)
                total_charge += hour_charge * perm
                hour.permittance = perm
            else:
                hour.permittance = 0.0
        self.model.input_hours = hours

    def _set_expected_charge(self, desired, peak, avg24) -> float:
        return (desired - self.total_charge) / (peak - avg24)

    async def async_initial_charge(self, avg24, peak) -> float:
        _avg24 = round((avg24 / 1000), 1)
        self.model.expected_hourly_charge = peak - _avg24
        total = 24 * (peak - _avg24)  # todo: fix 24 to be dynamic
        total -= len(self.non_hours) * (peak - _avg24)
        total -= sum(self.dynamic_caution_hours.values()) * (peak - _avg24)
        return total

    async def async_sum_charge(self, avg24, peak) -> float:
        total = 0
        for k in self.model.input_hours:
            total += (peak - avg24) * k.permittance
        return total

    def _service_caution_hours(self) -> dict:
        return {
            hp.dt: hp.permittance
            for hp in self.parent.model.hours_prices
            if not hp.passed and 0.0 < hp.permittance < 1.0
        }

    def _service_non_hours(self) -> list:
        return [
            hp.dt
            for hp in self.parent.model.hours_prices
            if not hp.passed and hp.permittance == 0.0
        ]

    async def async_setup(
        self,
        max_charge: float,
        non_hours: list | None = None,
        dynamic_caution_hours: dict | None = None,
        prices: list | None = None,
        prices_tomorrow: list | None = None,
    ) -> None:
        if max_charge == 0:
            self.active = False
            return
        hour = self.parent.dtmodel.hour
        dt = self.parent.dtmodel.dt
        _non_hours = self._service_non_hours() if non_hours is None else non_hours
        _dynamic_caution_hours = (
            self._service_caution_hours()
            if dynamic_caution_hours is None
            else dynamic_caution_hours
        )
        _prices = self.parent.model.prices_today if prices is None else prices
        _prices_tomorrow = (
            self.parent.model.prices_tomorrow
            if prices_tomorrow is None
            else prices_tomorrow
        )

        self.model.input_hours = self.parent.future_hours
        self.model.original_input_hours = self.model.input_hours.copy()
        self.active = True



from ....models.locale.enums.querytype import QueryType
from ....models.locale.enums.sum_types import SumTypes
from ....models.locale.enums.time_periods import TimePeriods
from ....models.locale.sumcounter import SumCounter
from ..locale_query import LocaleQuery

QUERYTYPES = {
    QueryType.AverageOfThreeHours: LocaleQuery(
        sum_type=SumTypes.Avg,
        time_calc=TimePeriods.Hourly,
        cycle=TimePeriods.Monthly,
        sum_counter=SumCounter(counter=3, groupby=TimePeriods.Hourly),
    ),
    QueryType.AverageOfThreeDays: LocaleQuery(
        sum_type=SumTypes.Avg,
        time_calc=TimePeriods.Hourly,
        cycle=TimePeriods.Monthly,
        sum_counter=SumCounter(counter=3, groupby=TimePeriods.Daily),
    ),
    QueryType.Max: LocaleQuery(
        sum_type=SumTypes.Max, time_calc=TimePeriods.Hourly, cycle=TimePeriods.Monthly
    ),
    QueryType.AverageOfFiveDays: LocaleQuery(
        sum_type=SumTypes.Avg,
        time_calc=TimePeriods.Hourly,
        cycle=TimePeriods.Monthly,
        sum_counter=SumCounter(counter=5, groupby=TimePeriods.Daily),
    ),
}



from abc import abstractmethod
from ...util import _convert_quarterly_minutes, async_convert_quarterly_minutes
from datetime import datetime
import logging
from ...models.phases import Phases
from ...models.const import (
    CURRENTS_ONEPHASE_1_16,
    CURRENTS_THREEPHASE_1_16,
    CURRENTS_ONEPHASE_1_32,
    CURRENTS_THREEPHASE_1_32,
)

_LOGGER = logging.getLogger(__name__)


ALGORITHM_INPUTS = {
    "stop_caution": [1.075, 0.0032, 0.7],
    "stop": [1.071, 0.00165, 0.8],
    "start_caution": [1.081, 0.0049, 0.4],
    "start": [1.066, 0.0045, 0.5],
}


async def async_calculate_algorithm_inputs(type: str, minute: int) -> float:
    inputs = ALGORITHM_INPUTS[type]
    return round((((minute + pow(inputs[0], minute)) * inputs[1]) + inputs[2]) * 100, 2)


def _calculate_algorithm_inputs(type: str, minute: int) -> float:
    """legacy. remove once peaqhvac is async."""
    inputs = ALGORITHM_INPUTS[type]
    return round((((minute + pow(inputs[0], minute)) * inputs[1]) + inputs[2]) * 100, 2)


class ThresholdBase:
    BASECURRENT = 6

    def __init__(self, hub):
        self._hub = hub
        self._phases = Phases.Unknown
        self._currents = {}

    @property
    def phases(self) -> str:
        return self._phases.name

    @property
    def currents(self) -> dict:
        return self._currents

    # @property
    # @abstractmethod
    # def allowedcurrent(self) -> int:
    #     pass

    def _setcurrentdict(self) -> dict:
        """only allow amps if user has set this value high enough"""
        if self._hub.chargertype.max_amps != 16:
            _threephase = {
                k: v
                for (k, v) in CURRENTS_THREEPHASE_1_32.items()
                if v <= self._hub.chargertype.max_amps
            }
            _onephase = {
                k: v
                for (k, v) in CURRENTS_ONEPHASE_1_32.items()
                if v <= self._hub.chargertype.max_amps
            }
        else:
            _threephase = CURRENTS_THREEPHASE_1_16
            _onephase = CURRENTS_ONEPHASE_1_16
        if hasattr(self._hub.sensors, "carpowersensor"):
            try:
                divid = int(self._hub.sensors.carpowersensor.value) / int(
                    self._hub.sensors.chargerobject_switch.current
                )
                if int(self._hub.sensors.carpowersensor.value) == 0:
                    self._phases = Phases.Unknown
                    ret = _threephase
                elif divid < 300:
                    self._phases = Phases.OnePhase
                    ret = _onephase
                else:
                    self._phases = Phases.ThreePhase
                    ret = _threephase
            except:
                _LOGGER.debug(
                    "Currents-dictionary: could not divide charger amps with charger power. Falling back to legacy-method."
                )
                if 0 < int(self._hub.sensors.carpowersensor.value) < 4000:
                    self._phases = Phases.OnePhase
                    ret = _onephase
                else:
                    self._phases = Phases.ThreePhase
                    ret = _threephase
        else:
            self._phases = Phases.ThreePhase
            ret = _threephase
        self._currents = ret
        return ret

    async def async_setcurrentdict(self) -> dict:
        """only allow amps if user has set this value high enough"""
        if self._hub.chargertype.max_amps != 16:
            _threephase = {
                k: v
                for (k, v) in CURRENTS_THREEPHASE_1_32.items()
                if v <= self._hub.chargertype.max_amps
            }
            _onephase = {
                k: v
                for (k, v) in CURRENTS_ONEPHASE_1_32.items()
                if v <= self._hub.chargertype.max_amps
            }
        else:
            _threephase = CURRENTS_THREEPHASE_1_16
            _onephase = CURRENTS_ONEPHASE_1_16
        if hasattr(self._hub.sensors, "carpowersensor"):
            try:
                divid = int(self._hub.sensors.carpowersensor.value) / int(
                    self._hub.sensors.chargerobject_switch.current
                )
                if int(self._hub.sensors.carpowersensor.value) == 0:
                    self._phases = Phases.Unknown
                    ret = _threephase
                elif divid < 300:
                    self._phases = Phases.OnePhase
                    ret = _onephase
                else:
                    self._phases = Phases.ThreePhase
                    ret = _threephase
            except:
                _LOGGER.debug(
                    "Currents-dictionary: could not divide charger amps with charger power. Falling back to legacy-method."
                )
                if 0 < int(self._hub.sensors.carpowersensor.value) < 4000:
                    self._phases = Phases.OnePhase
                    ret = _onephase
                else:
                    self._phases = Phases.ThreePhase
                    ret = _threephase
        else:
            self._phases = Phases.ThreePhase
            ret = _threephase
        self._currents = ret
        return ret

    @staticmethod
    async def async_stop(
        now_min: int | None = None,
        is_caution_hour: bool = False,
        is_quarterly: bool = False,
    ) -> float:
        _minute = now_min if now_min is not None else datetime.now().minute
        minute = await async_convert_quarterly_minutes(_minute, is_quarterly)
        if is_caution_hour and minute < 45:
            return await async_calculate_algorithm_inputs("stop_caution", minute)
        return await async_calculate_algorithm_inputs("stop", minute)

    @staticmethod
    async def async_start(
        now_min: int | None = None,
        is_caution_hour: bool = False,
        is_quarterly: bool = False,
    ) -> float:
        _minute = now_min if now_min is not None else datetime.now().minute
        minute = await async_convert_quarterly_minutes(_minute, is_quarterly)
        if is_caution_hour and minute < 45:
            return await async_calculate_algorithm_inputs("start_caution", minute)
        return await async_calculate_algorithm_inputs("start", minute)

    @staticmethod
    def base_allowed_current(
        now_min: int,
        now_sec: int,
        moving_avg: float,
        charger_enabled: bool,
        charger_done: bool,
        currents_dict: dict,
        total_energy: float,
        peak: float,
        is_quarterly: bool = False,
        power_canary_amp: int = -1,
    ) -> int:
        minute = _convert_quarterly_minutes(now_min, is_quarterly)
        ret = ThresholdBase.BASECURRENT
        if not charger_enabled or charger_done or moving_avg == 0:
            return ret
        currents = currents_dict
        minsec = minute * 60 + now_sec
        for key, value in currents.items():
            if (
                (((moving_avg + key) / 3600) * (3600 - minsec) + total_energy * 1000)
                / 1000
            ) < peak:
                ret = value
                break
        return min(ret, power_canary_amp) if power_canary_amp > -1 else ret

    @staticmethod
    async def async_base_allowed_current(
        now_min: int,
        now_sec: int,
        moving_avg: float,
        charger_enabled: bool,
        charger_done: bool,
        currents_dict: dict,
        total_energy: float,
        peak: float,
        is_quarterly: bool = False,
        power_canary_amp: int = -1,
    ) -> int:
        minute = await async_convert_quarterly_minutes(now_min, is_quarterly)
        ret = ThresholdBase.BASECURRENT
        if not charger_enabled or charger_done or moving_avg == 0:
            return ret
        currents = currents_dict
        minsec = minute * 60 + now_sec
        for key, value in currents.items():
            if (
                (((moving_avg + key) / 3600) * (3600 - minsec) + total_energy * 1000)
                / 1000
            ) < peak:
                ret = value
                break
        return min(ret, power_canary_amp) if power_canary_amp > -1 else ret

    # legacy below

    @property
    def stop(self) -> float:
        """legacy. remove once peaqhvac is async."""
        is_caution = (
            str(datetime.now().hour) in self._hub.hours.caution_hours
            if self._hub.options.price.price_aware is False
            else False
        )

        return ThresholdBase._stop(
            datetime.now().minute,
            is_caution,
        )

    @property
    def start(self) -> float:
        """legacy. remove once peaqhvac is async."""
        is_caution = (
            str(datetime.now().hour) in self._hub.hours.caution_hours
            if self._hub.options.price.price_aware is False
            else False
        )

        return ThresholdBase._start(
            datetime.now().minute,
            is_caution,
        )

    @staticmethod
    def _stop(now_min: int, is_caution_hour: bool, is_quarterly: bool = False) -> float:
        """legacy. remove once peaqhvac is async."""
        minute = _convert_quarterly_minutes(now_min, is_quarterly)
        if is_caution_hour and minute < 45:
            return _calculate_algorithm_inputs("stop_caution", minute)
        return _calculate_algorithm_inputs("stop", minute)

    @staticmethod
    def _start(
        now_min: int, is_caution_hour: bool, is_quarterly: bool = False
    ) -> float:
        """legacy. remove once peaqhvac is async."""
        minute = _convert_quarterly_minutes(now_min, is_quarterly)
        if is_caution_hour and minute < 45:
            return _calculate_algorithm_inputs("start_caution", minute)
        return _calculate_algorithm_inputs("start", minute)
